#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re

from .model import Reading, Position


def caesurae_together(position_specs, reward):
    def get_reward(meter: Meter, reading: Reading):
        for spec in position_specs:
            position = Position.after(spec[0], reading, spec[1], meter)
            if not position.word_boundary:
                return 0
        else:
            return reward
    return get_reward


def bridge(position_spec, reward):
    def get_reward(meter: Meter, reading: Reading):
        position = Position.after(position_spec[0], reading, meter,
                                  position_spec[1])
        if position.word_boundary:
            return 0
        else:
            return reward
    return get_reward


class Meter:

    def __init__(self, name: str, schema: str, conditions: list = None,
                 short_name: str = None):
        self.name = name
        self.schema = schema
        # Convert condition functions to instance-bound methods.
        self.conditions = ([cond.__get__(self) for cond in conditions]
                           if conditions else [])
        self.short_name = short_name

    def match_reading(self, reading: Reading):
        return re.match(self.schema, reading.get_schema())

    def get_rewards(self, reading: Reading):
        return sum(cond(reading) for cond in self.conditions)


AEOLIC_BASE = r'(?:(–)(–)|(–)(⏑)|(⏑)(–))'

ALL_METERS = {

    'Catalectic Dactylic Hexameter': Meter(
        'Catalectic Dactylic Hexameter',
        r'(–)(⏑⏑|–)(–)(⏑⏑|–)(–)(⏑⏑|–)(–)(⏑⏑|–)(–)(⏑⏑|–)(⏑|–)',
        conditions={
            caesurae_together([('mora', 6, 'Trithemimeral'),
                               ('mora', 14, 'Hephthemimeral')], 2),
            caesurae_together([('mora', 10, 'Penthemimeral')], 2),
            caesurae_together([('mora', 16, 'Bucolic Diaeresis')], 1),
            bridge(('mora', 15, 'Hermann’s Bridge'), 1)
        },
        short_name='6da‸'
    ),
    'Dactylic Pentameter': Meter(
        'Dactylic Pentameter',
        r'(–)(⏑⏑|–)(–)(⏑⏑|–)(–)(–)(⏑⏑)(–)(⏑⏑)(⏑|–)',
        conditions={
            caesurae_together([('mora', 5, 'Middle diaresis')], 2)
        },
        short_name='3da‸3da‸'
    ),
    'Iambic Trimeter': Meter(
        'Iambic Trimeter',
        r'(⏑|⏑⏑|–)(⏑⏑|–)(⏑)(⏑⏑|–)(⏑|⏑⏑|–)(⏑⏑|–)(⏑)(⏑⏑|–)(⏑|⏑⏑|–)(⏑⏑|–)(⏑)(⏑|–)',
        conditions={
            caesurae_together([('element', 4, 'After fourth element')], 1),
            caesurae_together([('element', 8, 'After eighth element')], 1),
        },
        short_name='3ia'
    ),
    'Iambic Senarius': Meter(
        'Iambic Senarius',
        r'(⏑|⏑⏑|–)(⏑⏑|–)(⏑|⏑⏑|–)(⏑⏑|–)(⏑|⏑⏑|–)(⏑⏑|–)(⏑|⏑⏑|–)(⏑⏑|–)(⏑|⏑⏑|–)(⏑⏑|–)(⏑)(⏑|–)',
        short_name='6ia'
    ),
    'Sapphic Hendecasyllable': Meter(
        'Sapphic Hendecasyllable',
        r'(–)(–|⏑)(–)(–|⏑)(–)(⏑)(⏑)(–)(⏑)(–)(⏑|–)',
        conditions={},
        short_name='sap hen'
    ),
    'Adoneus': Meter(
        'Adoneus',
        r'(–)(⏑⏑)(–)(⏑|–)',
        short_name='adoneus',
    ),
    'Phalaecian Hendecasyllable': Meter(
        'Phalaecian Hendecasyllable',
        AEOLIC_BASE + r'(–)(⏑)(⏑)(–)(⏑)(–)(⏑)(–)(⏑|–)',
        conditions={
            caesurae_together([('element', 6, 'After sixth element')], 1)
        },
        short_name='hen'
    ),
}
